#!/usr/bin/env python3
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Works with (amongst others):
- evohome (up to 12 zones)
- sundial (up to 2 zones)
- chronotherm (CM60xNG can do 4 zones)
- hometronics (16? zones)
- vision pro
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ramses_tx import Address, Command, Message, Packet  # noqa: F401

from .device import Device  # noqa: F401
from .gateway import Gateway  # noqa: F401
from .version import VERSION  # noqa: F401

from .const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    I_,
    RP,
    RQ,
    W_,
    Code,
)

if TYPE_CHECKING:
    from .const import IndexT, VerbT  # noqa: F401, pylint: disable=unused-import


_LOGGER = logging.getLogger(__name__)


class GracefulExit(SystemExit):
    code = 1



#!/usr/bin/env python3
"""RAMSES RF - exceptions above the packet/protocol/transport layer."""

from __future__ import annotations

from ramses_tx.exceptions import (  # noqa: F401
    PacketAddrSetInvalid,
    PacketInvalid,
    PacketPayloadInvalid,
    ProtocolError,
    RamsesException,
)


class _RamsesUpperError(RamsesException):
    """A failure in the upper layer (state/schema, gateway, bindings, schedule)."""


########################################################################################
# Errors above the protocol/transport layer, incl. message processing, state & schema


class BindingError(_RamsesUpperError):
    """An error occured when binding."""


class BindingFsmError(BindingError):
    """The binding FSM was/became inconsistent (this shouldn't happen)."""


class BindingFlowFailed(BindingError):
    """The binding failed due to a timeout or retry limit being exceeded."""


########################################################################################
# Errors above the protocol/transport layer, incl. message processing, state & schema


class ScheduleError(_RamsesUpperError):
    """An error occured when getting/setting a schedule."""


class ScheduleFsmError(BindingError):
    """The schedule FSM was/became inconsistent (this shouldn't happen)."""


class ScheduleFlowError(BindingError):
    """The get/set schedule failed due to a timeout or retry limit being exceeded."""


########################################################################################
# Errors above the protocol/transport layer, incl. message processing, state & schema


class ExpiredCallbackError(_RamsesUpperError):
    """Raised when the callback has expired."""


class SystemInconsistent(_RamsesUpperError):
    """Base class for exceptions in this module."""


class SystemSchemaInconsistent(SystemInconsistent):
    """Raised when the system state (usu. schema) is inconsistent."""

    HINT = "try restarting the client library"


class ForeignGatewayError(_RamsesUpperError):
    """Raised when a foreign gateway is detected.

    These devices may not be gateways (set a class), or belong to a neighbour (exclude
    via block_list/known_list), or should be allowed (known_list).
    """

    HINT = "consider enforcing a known_list"



#!/usr/bin/env python3

# TODO: Add assert_protocol_ready to VirtualRF factory (or in library?)

"""Test the Virtual RF library.

VirtualRF is used for testing.
"""

import asyncio

import pytest
import serial  # type: ignore[import-untyped]

from ramses_rf import Address, Code, Command, Gateway
from ramses_tx.transport import PortTransport
from tests_rf.virtual_rf import VirtualRf, rf_factory

# other constants
ASSERT_CYCLE_TIME = 0.001  # max_cycles_per_assert = max_sleep / ASSERT_CYCLE_TIME
DEFAULT_MAX_SLEEP = 1


GWY_CONFIG = {
    "config": {
        "disable_discovery": True,  # we're testing discovery here
        "enforce_known_list": False,
    },
}


SCHEMA_0 = {
    "orphans_hvac": ["40:000000"],
    "known_list": {"40:000000": {"class": "REM"}},
}

SCHEMA_1 = {
    "orphans_hvac": ["41:111111"],
    "known_list": {"41:111111": {"class": "FAN"}},
}


# ######################################################################################


async def assert_code_in_device_msgz(
    gwy: Gateway,
    dev_id: str,
    code: Code,
    max_sleep: int = DEFAULT_MAX_SLEEP,
    test_not: bool = False,
) -> None:
    """Fail if the device doesn't exist, or if it doesn't have the code in its DB."""

    for _ in range(int(max_sleep / ASSERT_CYCLE_TIME)):
        await asyncio.sleep(ASSERT_CYCLE_TIME)
        if ((dev := gwy.device_by_id.get(dev_id)) and (code in dev._msgz)) != test_not:
            break
    assert (
        (dev := gwy.device_by_id.get(dev_id)) and (code in dev._msgz)
    ) != test_not  # TODO: fix me


async def assert_devices(
    gwy: Gateway, devices: list[str], max_sleep: int = DEFAULT_MAX_SLEEP
) -> None:
    """Fail if the two sets of devices are not equal."""

    devices = [Address(d).id for d in devices]

    for _ in range(int(max_sleep / ASSERT_CYCLE_TIME)):
        await asyncio.sleep(ASSERT_CYCLE_TIME)
        if len(gwy.devices) == len(devices):
            break
    assert sorted(d.id for d in gwy.devices) == sorted(devices)


async def assert_this_pkt(
    transport: PortTransport, cmd: Command, max_sleep: int = DEFAULT_MAX_SLEEP
) -> None:
    """Check, at the transport layer, that the current packet is as expected."""
    for _ in range(int(max_sleep / ASSERT_CYCLE_TIME)):
        await asyncio.sleep(ASSERT_CYCLE_TIME)
        if transport._this_pkt and transport._this_pkt._frame == cmd._frame:
            break
    assert transport._this_pkt and transport._this_pkt._frame == cmd._frame


# ### TESTS ############################################################################


async def _test_virtual_rf_dev_disc(
    rf: VirtualRf, gwy_0: Gateway, gwy_1: Gateway
) -> None:
    """Check the virtual RF network behaves as expected (device discovery)."""

    ser_2 = serial.Serial(rf.ports[2])

    # TEST 0: Tx of fingerprint packet with one on/one off
    await gwy_0.start()
    assert gwy_0._protocol._transport

    await assert_devices(gwy_0, ["18:000000"])
    await assert_devices(gwy_1, [])

    await gwy_1.start()
    assert gwy_1._protocol._transport

    # NOTE: will pick up gwy 18:111111, since Foreign gwy detect has been removed
    await assert_devices(gwy_0, ["18:000000", "18:111111"])
    await assert_devices(gwy_1, ["18:111111"])

    # TEST 1: Tx to all from GWY /dev/pty/0 (NB: no RSSI)
    cmd = Command(" I --- 01:010000 --:------ 01:010000 1F09 003 0004B5")
    gwy_0.send_cmd(cmd)

    await assert_devices(gwy_0, ["01:010000", "18:000000", "18:111111"])
    await assert_devices(gwy_1, ["01:010000", "18:111111"])

    # TEST 2: Tx to all from non-GWY /dev/pty/2 (NB: no RSSI)
    cmd = Command(" I --- 01:011111 --:------ 01:011111 1F09 003 0004B5")
    ser_2.write(bytes(f"{cmd}\r\n".encode("ascii")))

    await assert_devices(gwy_0, ["01:010000", "01:011111", "18:000000", "18:111111"])
    await assert_devices(gwy_1, ["01:010000", "01:011111", "18:111111"])

    # TEST 3: Rx only by *only one* GWY (NB: needs RSSI)
    cmd = Command(" I --- 01:022222 --:------ 01:022222 1F09 003 0004B5")
    list(rf._port_to_object.values())[1].write(bytes(f"000 {cmd}\r\n".encode("ascii")))

    await assert_devices(gwy_0, ["01:010000", "01:011111", "18:000000", "18:111111"])
    await assert_devices(gwy_1, ["01:010000", "01:011111", "01:022222", "18:111111"])


async def _test_virtual_rf_pkt_flow(
    rf: VirtualRf, gwy_0: Gateway, gwy_1: Gateway
) -> None:
    """Check the virtual RF network behaves as expected (packet flow)."""

    # TEST 1:
    await assert_code_in_device_msgz(
        gwy_0, "01:022222", Code._1F09, max_sleep=0, test_not=True
    )  # device wont exist

    cmd = Command(" I --- 01:022222 --:------ 01:022222 1F09 003 0004B5")
    gwy_0.send_cmd(cmd, num_repeats=1)

    await assert_devices(gwy_0, ["01:022222", "18:000000", "18:111111", "40:000000"])
    await assert_code_in_device_msgz(gwy_0, "01:022222", Code._1F09)

    await assert_this_pkt(gwy_0._transport, cmd)
    await assert_this_pkt(gwy_1._transport, cmd)

    # TEST 2:
    await assert_code_in_device_msgz(
        gwy_0, "40:000000", Code._22F1, max_sleep=0, test_not=True
    )

    cmd = Command(" I --- 40:000000 --:------ 40:000000 22F1 003 000507")
    gwy_0.send_cmd(cmd, num_repeats=1)

    # await assert_code_in_device_msgz(gwy_0, "40:000000", Code._22F1)  # ?needs QoS

    await assert_this_pkt(gwy_0._transport, cmd)
    await assert_this_pkt(gwy_1._transport, cmd)

    await assert_devices(gwy_0, ["01:022222", "18:000000", "18:111111", "40:000000"])
    await assert_devices(gwy_1, ["01:022222", "18:111111", "40:000000", "41:111111"])


# NOTE: does not use factory
@pytest.mark.xdist_group(name="virt_serial")
async def test_virtual_rf_dev_disc() -> None:
    """Check the virtual RF network behaves as expected (device discovery)."""

    rf = VirtualRf(3)

    gwy_0: Gateway = None  # type: ignore[assignment]
    gwy_1: Gateway = None  # type: ignore[assignment]

    try:
        rf.set_gateway(rf.ports[0], "18:000000")
        gwy_0 = Gateway(rf.ports[0], **GWY_CONFIG)  # type: ignore[arg-type]
        await assert_devices(gwy_0, [])

        rf.set_gateway(rf.ports[1], "18:111111")
        gwy_1 = Gateway(rf.ports[1], **GWY_CONFIG)  # type: ignore[arg-type]
        await assert_devices(gwy_1, [])

        await _test_virtual_rf_dev_disc(rf, gwy_0, gwy_1)

    finally:
        if gwy_0:
            await gwy_0.stop()
        if gwy_1:
            await gwy_1.stop()
        await rf.stop()


# NOTE: uses factory
@pytest.mark.xdist_group(name="virt_serial")
async def test_virtual_rf_pkt_flow() -> None:
    """Check the virtual RF network behaves as expected (packet flow)."""

    rf: VirtualRf = None  # type: ignore[assignment]

    try:
        rf, (gwy_0, gwy_1) = await rf_factory(
            [GWY_CONFIG | SCHEMA_0, GWY_CONFIG | SCHEMA_1]
        )

        assert gwy_0._protocol._transport
        # NOTE: will pick up gwy 18:111111, since Foreign gwy detect has been removed
        await assert_devices(gwy_0, ["18:000000", "18:111111", "40:000000"])

        assert gwy_1._protocol._transport
        await assert_devices(gwy_1, ["18:111111", "41:111111"])

        await _test_virtual_rf_pkt_flow(rf, gwy_0, gwy_1)

    finally:
        if rf:
            await gwy_0.stop()
            await gwy_1.stop()
            await rf.stop()



#!/usr/bin/env python3
"""RAMSES RF - Test the Schema processor."""

import voluptuous as vol

from ramses_rf.schemas import (
    SCH_TCS_DHW,
    SCH_TCS_ZONES_ZON,
    SZ_ACTUATORS,
    SZ_CLASS,
    SZ_DHW_VALVE,
    SZ_HTG_VALVE,
    SZ_SENSOR,
)
from tests.helpers import assert_raises


def test_dhw_schema() -> None:
    """Test the DHW schema.

    dhw:
        sensor: 07:777777
        hotwater_valve: 13:111111
        heating_valve: 13:222222
    """

    for dict_ in (
        SCH_TCS_DHW({}),
        {SZ_SENSOR: None, SZ_DHW_VALVE: None, SZ_HTG_VALVE: None},
    ):
        assert dict_ == {SZ_SENSOR: None, "hotwater_valve": None, "heating_valve": None}

    for key in (SZ_SENSOR, SZ_DHW_VALVE, SZ_HTG_VALVE):
        assert_raises(vol.error.MultipleInvalid, SCH_TCS_DHW, {key: "99:000000"})
        assert SCH_TCS_DHW({key: None}) == {
            SZ_SENSOR: None,
            SZ_DHW_VALVE: None,
            SZ_HTG_VALVE: None,
        }


def test_zone_schema() -> None:
    """Test the zone schema.

    '01':
        class: radiator_valve
        sensor: 22:032844
        actuators:
        - 04:111111
        - 04:222222
    """

    for dict_ in (
        SCH_TCS_ZONES_ZON({}),
        {SZ_CLASS: None, SZ_SENSOR: None, SZ_ACTUATORS: []},
    ):
        assert dict_ == {SZ_CLASS: None, SZ_SENSOR: None, SZ_ACTUATORS: []}

    for key in (SZ_CLASS, SZ_SENSOR):
        assert_raises(vol.error.MultipleInvalid, SCH_TCS_ZONES_ZON, {key: "99:000000"})
        assert SCH_TCS_ZONES_ZON({key: None}) == {
            SZ_CLASS: None,
            SZ_SENSOR: None,
            SZ_ACTUATORS: [],
        }

    assert SCH_TCS_ZONES_ZON({SZ_CLASS: "radiator_valve"}) == {
        SZ_CLASS: "radiator_valve",
        SZ_SENSOR: None,
        SZ_ACTUATORS: [],
    }

    assert SCH_TCS_ZONES_ZON({SZ_SENSOR: "34:111111"}) == {
        SZ_CLASS: None,
        SZ_ACTUATORS: [],
        SZ_SENSOR: "34:111111",
    }

    for val in (
        None,
        "_invalid_",
        "13:111111",
    ):  # NOTE: should be a *list* of device_ids
        assert_raises(vol.error.MultipleInvalid, SCH_TCS_ZONES_ZON, {SZ_ACTUATORS: val})

    for xxx in ([], ["13:111111"], ["13:222222", "13:111111"]):
        assert SCH_TCS_ZONES_ZON({SZ_ACTUATORS: xxx}) == {
            SZ_CLASS: None,
            SZ_ACTUATORS: xxx,
            SZ_SENSOR: None,
        }



#!/usr/bin/env python3
"""RAMSES RF - Test the RAMSES II schema."""

from ramses_rf import RQ
from ramses_rf.device.heat import HEAT_CLASS_BY_SLUG
from ramses_rf.device.hvac import HVAC_CLASS_BY_SLUG
from ramses_tx.const import Code, DevType
from ramses_tx.ramses import (
    _DEV_KLASSES_HEAT,
    _DEV_KLASSES_HVAC,
    _HVAC_VC_PAIR_BY_CLASS,
    CODE_IDX_ARE_COMPLEX,
    CODE_IDX_ARE_NONE,
    CODE_IDX_ARE_SIMPLE,
    CODES_SCHEMA,
    HVAC_KLASS_BY_VC_PAIR,
    RQ_NO_PAYLOAD,
)


def test_code_counts() -> None:
    """All known command codes should be in the schema & vice-versa."""

    # assert len(Code) == len(CODES_SCHEMA)
    assert not [c for c in CODES_SCHEMA if c not in Code]
    assert not [c for c in Code if c not in CODES_SCHEMA]


def test_verb_code_pairs() -> None:
    """Verb/code pairs are used to detect HVAC device classes: they should be unique."""

    assert len(HVAC_KLASS_BY_VC_PAIR) == (
        sum(len(v) for v in _HVAC_VC_PAIR_BY_CLASS.values())
    ), "Coding error: There is a duplicate verb/code pair"


def test_device_heat_slugs() -> None:
    """Every Heat device slug should have an entry in it domain's _DEV_KLASSES_*."""

    assert not [s for s in _DEV_KLASSES_HEAT if s not in HEAT_CLASS_BY_SLUG]
    assert not [
        s for s in HEAT_CLASS_BY_SLUG if s not in _DEV_KLASSES_HEAT and s != DevType.HEA
    ]


def test_device_hvac_slugs() -> None:
    """Every HVAC device slug should have an entry in it domain's _DEV_KLASSES_*."""

    assert not [s for s in _DEV_KLASSES_HVAC if s not in HVAC_CLASS_BY_SLUG]
    assert not [
        s for s in HVAC_CLASS_BY_SLUG if s not in _DEV_KLASSES_HVAC and s != DevType.HVC
    ]


def assert_codes_idx_mutex(mutex_list: set, other_list: set) -> None:
    """Assert the two lists are mutually exclusive."""

    codes = sorted(c for c in mutex_list if c in other_list)
    assert not codes


def test_codes_idx_mutex() -> None:
    """Every code should be in one of the three CODE_IDX_* constants."""

    codes_idx_all = CODE_IDX_ARE_COMPLEX | CODE_IDX_ARE_NONE | CODE_IDX_ARE_SIMPLE
    assert not [c for c in CODES_SCHEMA if c not in codes_idx_all]


def test_codes_idx_complex_mutex() -> None:
    """The three CODE_IDX_* constants should be mutally exclusive."""

    assert_codes_idx_mutex(
        CODE_IDX_ARE_COMPLEX, CODE_IDX_ARE_NONE | CODE_IDX_ARE_SIMPLE
    )


def test_codes_idx_none_mutex() -> None:
    """The three CODE_IDX_* constants should be mutally exclusive."""

    assert_codes_idx_mutex(
        CODE_IDX_ARE_NONE, CODE_IDX_ARE_SIMPLE | CODE_IDX_ARE_COMPLEX
    )


def test_codes_idx_simple_mutex() -> None:
    """The three CODE_IDX_* constants should be mutally exclusive."""

    assert_codes_idx_mutex(
        CODE_IDX_ARE_SIMPLE, CODE_IDX_ARE_NONE | CODE_IDX_ARE_COMPLEX
    )


def test_codes_mutex() -> None:
    assert_codes_idx_mutex(RQ_IDX_ONLY, CODE_IDX_ARE_NONE)


RQ_IDX_NONE = [k for k, v in CODES_SCHEMA.items() if v.get(RQ, "")[:3] == "^00"]
RQ_IDX_ONLY = [
    k
    for k, v in CODES_SCHEMA.items()
    if k not in RQ_NO_PAYLOAD and (v.get(RQ) in (r"^0[0-9A-F]00$", r"^0[0-9A-F](00)?$"))
]
RQ_IDX_UNKNOWN = [
    k
    for k, v in CODES_SCHEMA.items()
    if k not in RQ_NO_PAYLOAD + RQ_IDX_ONLY and RQ in v
]



#!/usr/bin/env python3
"""Fixtures for testing."""

import logging
import os
from collections.abc import AsyncGenerator
from typing import Final, NoReturn, TypeAlias, TypedDict
from unittest.mock import patch

import pytest
import serial as ser  # type: ignore[import-untyped]
from serial.tools.list_ports import comports  # type: ignore[import-untyped]

from ramses_rf import Gateway
from ramses_rf.device import HgiGateway
from ramses_tx import exceptions as exc
from ramses_tx.address import HGI_DEVICE_ID
from ramses_tx.schemas import DeviceIdT
from tests_rf.virtual_rf import HgiFwTypes, VirtualRf

#
PortStrT: TypeAlias = str

SZ_GWY_CONFIG: Final = "gwy_config"
SZ_GWY_DEV_ID: Final = "gwy_dev_id"


class _ConfigDictT(TypedDict):
    disable_discovery: bool
    disable_qos: bool
    enforce_known_list: bool


class _GwyConfigDictT(TypedDict):
    config: _ConfigDictT
    known_list: dict[DeviceIdT, dict[str, bool]]


_LOGGER = logging.getLogger(__name__)


IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"
# set -x GITHUB_ACTIONS true
# set -u GITHUB_ACTIONS

_global_failed_ports: list[str] = []


#######################################################################################

# pytestmark = pytest.mark.asyncio(scope="function")  # needed?


@pytest.fixture(autouse=True)
def patches_for_tests(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("ramses_tx.protocol._DBG_DISABLE_IMPERSONATION_ALERTS", True)
    monkeypatch.setattr("ramses_tx.transport._DBG_DISABLE_DUTY_CYCLE_LIMIT", True)
    monkeypatch.setattr("ramses_tx.transport.MINIMUM_WRITE_GAP", 0)


# TODO: add teardown to cleanup orphan MessageIndex thread
# @pytest.fixture(scope="session", autouse=True)
# def close_timer_threads(request: pytest.FixtureRequest) -> None:
#     import threading

#     def finalize() -> None:
#         running_timer_threads = [
#             thread
#             for thread in threading.enumerate()
#             if isinstance(thread, threading.Timer)
#         ]
#         for timer in running_timer_threads:
#             timer.cancel()

#     request.addfinalizer(finalize)


#######################################################################################


@pytest.fixture()
async def rf() -> AsyncGenerator[VirtualRf, None]:
    """Utilize a virtual evofw3-compatible gateway."""

    rf = VirtualRf(2)

    try:
        yield rf
    finally:
        await rf.stop()


@pytest.fixture()
def fake_evofw3_port(request: pytest.FixtureRequest, rf: VirtualRf) -> PortStrT | None:
    """Utilize a virtual evofw3-compatible gateway."""

    gwy_dev_id: DeviceIdT = request.getfixturevalue(SZ_GWY_DEV_ID)

    rf.set_gateway(rf.ports[0], gwy_dev_id, fw_type=HgiFwTypes.EVOFW3)

    # with patch("ramses_tx.transport.comports", rf.comports):
    return rf.ports[0]


@pytest.fixture()
def fake_ti3410_port(request: pytest.FixtureRequest, rf: VirtualRf) -> PortStrT | None:
    """Utilize a virtual HGI80-compatible gateway."""

    gwy_dev_id: DeviceIdT = request.getfixturevalue(SZ_GWY_DEV_ID)

    rf.set_gateway(rf.ports[0], gwy_dev_id, fw_type=HgiFwTypes.HGI_80)

    # with patch("ramses_tx.transport.comports", rf.comports):
    return rf.ports[0]


@pytest.fixture()  # TODO: remove HACK, below
async def mqtt_evofw3_port() -> PortStrT:
    """Utilize an actual evofw3-compatible gateway."""

    if IN_GITHUB_ACTIONS:  # replace with your condition
        pytest.skip("Test doesn't work without H/W")

    # TODO: add a test & pytest.skip() if no MQTT broker is available

    return "mqtt://mqtt_username:mqtt_passw0rd@127.0.0.1"


@pytest.fixture()  # TODO: remove HACK, below
async def real_evofw3_port() -> PortStrT | NoReturn:
    """Utilize an actual evofw3-compatible gateway."""

    port_names: list[PortStrT] = [
        p.device for p in comports() if p.product and "evofw3" in p.product
    ]

    if port_names:
        return port_names[0]

    if port_names := [
        p.device for p in comports() if p.name[:6] == "ttyACM"
    ]:  # HACK: evofw3-esp
        _LOGGER.warning(f"Assuming {port_names[0]} is evofw3-compatible")
        return port_names[0]

    pytest.skip("No evofw3-based gateway device found")


@pytest.fixture()
async def real_ti3410_port() -> PortStrT | NoReturn:
    """Utilize an actual HGI80-compatible gateway."""

    port_names: list[PortStrT] = [
        p.device for p in comports() if p.product and "TUSB3410" in p.product
    ]

    if port_names:
        return port_names[0]

    pytest.skip("No ti3410-based gateway device found")


#######################################################################################


async def _gateway(gwy_port: PortStrT, gwy_config: _GwyConfigDictT) -> Gateway:
    """Instantiate a gateway."""

    gwy = Gateway(gwy_port, **gwy_config)

    assert gwy.hgi is None and gwy.devices == []

    await gwy.start()
    return gwy


async def _fake_gateway(
    gwy_port: PortStrT, gwy_config: _GwyConfigDictT, rf: VirtualRf
) -> Gateway:
    """Wrapper to instantiate a virtual gateway."""

    with patch("ramses_tx.transport.comports", rf.comports):
        gwy = await _gateway(gwy_port, gwy_config)

    assert gwy._transport  # mypy
    gwy._transport._extra["virtual_rf"] = rf
    return gwy


async def _real_gateway(gwy_port: PortStrT, gwy_config: _GwyConfigDictT) -> Gateway:
    """Wrapper to instantiate a physical gateway."""

    global _global_failed_ports

    if gwy_port in _global_failed_ports:
        pytest.skip(f"Port {gwy_port} previously failed")

    try:
        return await _gateway(gwy_port, gwy_config)
    except (ser.SerialException, exc.TransportError) as err:
        _global_failed_ports.append(gwy_port)
        pytest.xfail(str(err))  # not skip, as we had determined port exists elsewhere


@pytest.fixture()
async def fake_evofw3(
    fake_evofw3_port: PortStrT, request: pytest.FixtureRequest, rf: VirtualRf
) -> AsyncGenerator[Gateway, None]:
    """Utilize a virtual evofw3-compatible gateway (discovered by fake_evofw3_port)."""

    gwy_config: _GwyConfigDictT = request.getfixturevalue(SZ_GWY_CONFIG)
    gwy_dev_id: DeviceIdT = request.getfixturevalue(SZ_GWY_DEV_ID)

    gwy = await _fake_gateway(fake_evofw3_port, gwy_config, rf)

    assert isinstance(gwy.hgi, HgiGateway) and gwy.hgi.id == gwy_dev_id
    assert gwy._protocol._is_evofw3 is True

    try:
        yield gwy
    finally:
        await gwy.stop()


@pytest.fixture()
async def fake_ti3410(
    fake_ti3410_port: PortStrT, request: pytest.FixtureRequest, rf: VirtualRf
) -> AsyncGenerator[Gateway, None]:
    """Utilize a virtual HGI80-compatible gateway (discovered by fake_ti3410_port)."""

    gwy_config: _GwyConfigDictT = request.getfixturevalue(SZ_GWY_CONFIG)
    gwy_dev_id: DeviceIdT = request.getfixturevalue(SZ_GWY_DEV_ID)

    gwy = await _fake_gateway(fake_ti3410_port, gwy_config, rf)

    assert isinstance(gwy.hgi, HgiGateway) and gwy.hgi.id == gwy_dev_id
    assert gwy._protocol._is_evofw3 is False

    try:
        yield gwy
    finally:
        await gwy.stop()


@pytest.fixture()
async def mqtt_evofw3(
    mqtt_evofw3_port: PortStrT, request: pytest.FixtureRequest
) -> AsyncGenerator[Gateway, None]:
    """Utilize an actual evofw3-compatible gateway (discovered by mqtt_evofw3_port)."""

    gwy_config: _GwyConfigDictT = request.getfixturevalue(SZ_GWY_CONFIG)

    gwy = await _real_gateway(mqtt_evofw3_port, gwy_config)

    gwy.get_device(gwy._protocol.hgi_id)  # HACK: not instantiated: no puzzle pkts sent

    assert isinstance(gwy.hgi, HgiGateway) and gwy.hgi.id not in (None, HGI_DEVICE_ID)
    assert gwy._protocol._is_evofw3 is True

    try:
        yield gwy
    finally:
        await gwy.stop()


@pytest.fixture()
async def real_evofw3(
    real_evofw3_port: PortStrT, request: pytest.FixtureRequest
) -> AsyncGenerator[Gateway, None]:
    """Utilize an actual evofw3-compatible gateway (discovered by real_evofw3_port)."""

    gwy_config: _GwyConfigDictT = request.getfixturevalue(SZ_GWY_CONFIG)

    gwy = await _real_gateway(real_evofw3_port, gwy_config)

    assert isinstance(gwy.hgi, HgiGateway) and gwy.hgi.id not in (None, HGI_DEVICE_ID)
    assert gwy._protocol._is_evofw3 is True

    try:
        yield gwy
    finally:
        await gwy.stop()


@pytest.fixture()
async def real_ti3410(
    real_ti3410_port: PortStrT, request: pytest.FixtureRequest
) -> AsyncGenerator[Gateway, None]:
    """Utilize an actual HGI80-compatible gateway (discovered by real_ti3410_port)."""

    gwy_config: _GwyConfigDictT = request.getfixturevalue(SZ_GWY_CONFIG)

    gwy = await _real_gateway(real_ti3410_port, gwy_config)

    assert isinstance(gwy.hgi, HgiGateway) and gwy.hgi.id not in (None, HGI_DEVICE_ID)
    assert gwy._protocol._is_evofw3 is False

    try:
        yield gwy
    finally:
        await gwy.stop()



#!/usr/bin/env python3
"""RAMSES RF - Test the binding protocol with a virtual RF

NB: This test will likely fail with pytest -n x, because of the protocol's throttle
limits.
"""

import asyncio
import random
from collections.abc import AsyncGenerator, Awaitable
from datetime import datetime as dt

import pytest
import serial  # type: ignore[import-untyped]

from ramses_rf import Command, Message, Packet
from ramses_tx import exceptions as exc
from ramses_tx.protocol import PortProtocol, ReadProtocol, protocol_factory
from ramses_tx.protocol_fsm import (
    Inactive,
    IsInIdle,
    ProtocolContext,
    WantEcho,
    WantRply,
    _ProtocolStateT,
)
from ramses_tx.transport import transport_factory
from ramses_tx.typing import QosParams

from .virtual_rf import VirtualRf

# patched constants
DEFAULT_MAX_RETRIES = 0  # #                ramses_tx.protocol
MAX_DUTY_CYCLE = 1.0  # #                   ramses_tx.protocol

# other constants
CALL_LATER_DELAY = 0.001  # FIXME: this is hardware-specific

ASSERT_CYCLE_TIME = 0.0005  # max_cycles_per_assert = max_sleep / ASSERT_CYCLE_TIME
DEFAULT_MAX_SLEEP = 0.1


# TODO: better handling than AttributeError for this...
# Command("RQ --- 18:111111 01:222222 --:------ 12B0 003 07")

II_CMD_STR_0 = " I --- 01:006056 --:------ 01:006056 1F09 003 0005C8"
II_CMD_0 = Command(II_CMD_STR_0)
II_PKT_0 = Packet(dt.now(), f"... {II_CMD_STR_0}")

# TIP: using 18:000730 as the source will prevent impersonation alerts

RQ_CMD_STR_0 = "RQ --- 18:000730 01:222222 --:------ 12B0 001 00"
RP_CMD_STR_0 = "RP --- 01:222222 18:000730 --:------ 12B0 003 000000"

RQ_CMD_0 = Command(RQ_CMD_STR_0)
RQ_PKT_0 = Packet(dt.now(), f"... {RQ_CMD_STR_0}")
RP_PKT_0 = Packet(dt.now(), f"... {RP_CMD_STR_0}")

RQ_CMD_STR_1 = "RQ --- 18:000730 01:222222 --:------ 12B0 001 01"
RP_CMD_STR_1 = "RP --- 01:222222 18:000730 --:------ 12B0 003 010000"

RQ_CMD_1 = Command(RQ_CMD_STR_1)
RQ_PKT_1 = Packet(dt.now(), f"... {RQ_CMD_STR_1}")
RP_PKT_1 = Packet(dt.now(), f"... {RP_CMD_STR_1}")


# ### FIXTURES #########################################################################


@pytest.fixture()
async def protocol(rf: VirtualRf) -> AsyncGenerator[PortProtocol, None]:
    def _msg_handler(msg: Message) -> None:
        pass

    protocol = protocol_factory(_msg_handler)

    # These values should be asserted as needed for subsequent tests
    assert isinstance(protocol, PortProtocol)  # mypy
    assert isinstance(protocol._context, ProtocolContext)  # mypy

    protocol._disable_qos = False  # HACK: needed for tests to succeed (default: None?)

    assert protocol._context.echo_timeout == 0.5
    assert protocol._context.reply_timeout == 0.5
    assert protocol._context.SEND_TIMEOUT_LIMIT == 20.0

    await assert_protocol_state(protocol, Inactive, max_sleep=0)

    transport = await transport_factory(protocol, port_name=rf.ports[0], port_config={})
    transport._extra["virtual_rf"] = rf  # injected to aid any debugging

    await assert_protocol_state(protocol, IsInIdle, max_sleep=0)

    try:
        yield protocol

    except serial.SerialException as err:
        transport._close(exc=err)
        raise

    except (AssertionError, asyncio.InvalidStateError, TimeoutError):
        transport.close()
        raise

    else:
        await assert_protocol_state(protocol, IsInIdle)
        transport.close()

    finally:
        await assert_protocol_state(protocol, Inactive, max_sleep=0.1)
        await rf.stop()


# ######################################################################################


async def assert_protocol_state(
    protocol: PortProtocol | ReadProtocol,
    expected_state: type[_ProtocolStateT],
    max_sleep: float = DEFAULT_MAX_SLEEP,
) -> None:
    assert isinstance(protocol, PortProtocol)  # mypy
    assert isinstance(protocol._context, ProtocolContext)  # mypy

    for _ in range(int(max_sleep / ASSERT_CYCLE_TIME)):
        await asyncio.sleep(ASSERT_CYCLE_TIME)
        if isinstance(protocol._context.state, expected_state):
            break
    assert isinstance(protocol._context.state, expected_state)


def assert_protocol_state_detail(
    protocol: PortProtocol, cmd: Command | None, num_sends: int
) -> None:
    assert isinstance(protocol._context, ProtocolContext)  # mypy

    assert protocol._context.state.cmd_sent == cmd
    assert protocol._context._cmd_tx_count == num_sends
    assert bool(cmd) is isinstance(protocol._context.state, WantEcho | WantRply)


async def async_pkt_received(  # type: ignore[no-any-unimported]
    protocol: PortProtocol,
    pkt: Packet,
    method: int = 0,
    ser: None | serial.Serial = None,
) -> None:
    # await assert_protocol_state(protocol, ProtocolState.IDLE, max_sleep=0)
    # assert_state_temp(protocol, None, 0)

    if method == 0:
        protocol.pkt_received(pkt)
        return

    if method == 1:
        protocol._loop.call_soon(protocol.pkt_received, pkt)
        return

    assert ser is not None
    frame = bytes(str(pkt).encode("ascii")) + b"\r\n"

    if method == 2:
        ser.write(frame)
    elif method == 3:
        protocol._loop.call_soon(ser.write, frame)
    else:
        protocol._loop.call_later(0.001, ser.write, frame)

    # await assert_protocol_state(protocol, ProtocolState.IDLE, max_sleep=0)
    # assert_state_temp(protocol, None, 0)


# ### TESTS ############################################################################


async def _test_flow_30x(protocol: PortProtocol) -> None:
    # STEP 0: Setup...
    rf: VirtualRf = protocol._transport._extra["virtual_rf"]
    ser = serial.Serial(rf.ports[1])

    qos = QosParams(wait_for_reply=True)

    # STEP 1: Send an I cmd (no reply)...
    task = rf._loop.create_task(protocol._send_cmd(II_CMD_0, qos=qos), name="send_1")
    assert await task == II_CMD_0  # no reply pkt expected

    # STEP 2: Send an RQ cmd, then receive the corresponding RP pkt...
    task = rf._loop.create_task(protocol._send_cmd(RQ_CMD_0, qos=qos), name="send_2")
    protocol._loop.call_later(
        CALL_LATER_DELAY, ser.write, bytes(str(RP_PKT_0).encode("ascii")) + b"\r\n"
    )
    assert await task == RP_PKT_0

    # STEP 3: Send an I cmd (no reply) *twice*...
    task = rf._loop.create_task(protocol._send_cmd(II_CMD_0, qos=qos), name="send_3A")
    assert await task == II_CMD_0  # no reply pkt expected

    task = rf._loop.create_task(protocol._send_cmd(II_CMD_0, qos=qos), name="send_3B")
    assert await task == II_CMD_0  # no reply pkt expected

    # STEP 4: Send an RQ cmd, then receive the corresponding RP pkt...
    task = rf._loop.create_task(protocol._send_cmd(RQ_CMD_1, qos=qos), name="send_4A")
    # sk = rf._loop.create_task(protocol._send_cmd(RQ_CMD_1, qos=qos), name="send_4B")

    # TODO: make these deterministic so ser replies *only after* it receives cmd
    protocol._loop.call_later(
        CALL_LATER_DELAY, ser.write, bytes(str(RP_PKT_0).encode("ascii")) + b"\r\n"
    )
    protocol._loop.call_later(
        CALL_LATER_DELAY, ser.write, bytes(str(RP_PKT_1).encode("ascii")) + b"\r\n"
    )

    assert await task == RP_PKT_1


async def _test_flow_401(protocol: PortProtocol) -> None:
    qos = QosParams(wait_for_reply=False)

    numbers = list(range(24))
    tasks = {}

    for i in numbers:
        cmd = Command.put_sensor_temp("03:123456", i)
        tasks[i] = protocol._loop.create_task(protocol._send_cmd(cmd, qos=qos))

    assert await asyncio.gather(*tasks.values())

    for i in numbers:
        pkt = tasks[i].result()
        assert pkt == Command.put_sensor_temp("03:123456", i)


async def _test_flow_402(protocol: PortProtocol) -> None:
    qos = QosParams(wait_for_reply=False)

    numbers = list(range(24))
    tasks = {}

    for i in numbers:
        cmd = Command.put_sensor_temp("03:123456", i)
        tasks[i] = protocol._loop.create_task(protocol._send_cmd(cmd, qos=qos))

    random.shuffle(numbers)

    for i in numbers:
        pkt = await tasks[i]
        assert pkt == Command.put_sensor_temp("03:123456", i)


async def _test_flow_qos_helper(
    send_cmd_coro: Awaitable, will_fail: bool = False
) -> None:
    try:
        _ = await send_cmd_coro
    except exc.ProtocolSendFailed:
        pass
    else:
        assert False, f"Had expected {exc.ProtocolSendFailed}"


async def _test_flow_60x(protocol: PortProtocol, num_cmds: int = 1) -> None:
    #
    # Setup...
    tasks = []
    for idx in range(num_cmds):
        cmd = Command.get_zone_temp("01:123456", f"{idx:02X}")
        coro = protocol._send_cmd(cmd, qos=QosParams(wait_for_reply=False))
        tasks.append(protocol._loop.create_task(coro, name=f"cmd_{idx:02X}"))

    assert await asyncio.gather(*tasks)


async def _test_flow_qos(protocol: PortProtocol) -> None:
    assert isinstance(protocol._context, ProtocolContext)  # mypy

    # HACK: to reduce test time
    protocol._context.SEND_TIMEOUT_LIMIT = 0.01
    protocol._context.max_retry_limit = 0

    #
    # ### Simple test for an I (does not expect any reply)...

    cmd = Command.put_sensor_temp("03:000111", 19.5)
    pkt = await protocol._send_cmd(cmd)  # qos == QosParams()
    assert pkt == cmd, "Should be echo as there's no reply to wait for"

    cmd = Command.put_sensor_temp("03:000222", 19.5)
    pkt = await protocol._send_cmd(cmd, qos=None)  # qos == QosParams()
    assert pkt == cmd, "Should be echo as there's no reply to wait for"

    cmd = Command.put_sensor_temp("03:000333", 19.5)
    pkt = await protocol._send_cmd(cmd, qos=QosParams())
    assert pkt == cmd, "Should be echo as there's no reply to wait for"

    cmd = Command.put_sensor_temp("03:000444", 19.5)
    pkt = await protocol._send_cmd(cmd, qos=QosParams(wait_for_reply=None))
    assert pkt == cmd, "should be echo as there is no wait_for_reply"

    cmd = Command.put_sensor_temp("03:000555", 19.5)
    pkt = await protocol._send_cmd(cmd, qos=QosParams(wait_for_reply=False))
    assert pkt == cmd, "should be echo as there is no wait_for_reply"

    cmd = Command.put_sensor_temp("03:000666", 19.5)
    pkt = await protocol._send_cmd(cmd, qos=QosParams(wait_for_reply=True))
    assert pkt == cmd, "Should be echo as there's no reply to wait for"

    # # ### Simple test for an RQ (expects an RP)...

    cmd = Command.get_system_time("01:000111")
    pkt = await protocol._send_cmd(cmd)
    assert pkt == cmd, "Should be echo as there's no reply to wait for"

    cmd = Command.get_system_time("01:000222")
    pkt = await protocol._send_cmd(cmd, qos=None)
    assert pkt == cmd, "Should be echo as there's no reply to wait for"

    cmd = Command.get_system_time("01:000333")
    pkt = await protocol._send_cmd(cmd, qos=QosParams())
    assert pkt == cmd, "Should be echo as there's no reply to wait for"

    cmd = Command.get_system_time("01:000444")
    pkt = await protocol._send_cmd(cmd, qos=QosParams(wait_for_reply=None))
    assert pkt == cmd, "Should be echo as there is no wait_for_reply"

    cmd = Command.get_system_time("01:000555")
    pkt = await protocol._send_cmd(cmd, qos=QosParams(wait_for_reply=False))
    assert pkt == cmd, "Should be echo as there is no wait_for_reply"

    cmd = Command.get_system_time("01:000666")
    coro = protocol._send_cmd(cmd, qos=QosParams(wait_for_reply=True, timeout=0.05))
    await _test_flow_qos_helper(coro)

    # # ### Simple test for an I (does not expect any reply)...

    cmd = Command.put_sensor_temp("03:000999", 19.5)
    pkt = await protocol._send_cmd(cmd)
    assert pkt == cmd


# ######################################################################################


@pytest.mark.xdist_group(name="virt_serial")
async def test_flow_300(protocol: PortProtocol) -> None:
    """Check state change of RQ/I/RQ cmds using protocol methods."""
    await _test_flow_30x(protocol)


@pytest.mark.xdist_group(name="virt_serial")
async def test_flow_401(protocol: PortProtocol) -> None:
    """Throw a bunch of commands in a random order, and see that all are echo'd."""
    await _test_flow_401(protocol)


@pytest.mark.xdist_group(name="virt_serial")
async def test_flow_402(protocol: PortProtocol) -> None:
    """Throw a bunch of commands in a random order, and see that all are echo'd."""
    await _test_flow_402(protocol)


@pytest.mark.xdist_group(name="virt_serial")
async def test_flow_601(protocol: PortProtocol) -> None:
    """Check the wait_for_reply kwarg."""
    await _test_flow_60x(protocol)


@pytest.mark.xdist_group(name="virt_serial")
async def test_flow_602(protocol: PortProtocol) -> None:
    """Check the wait_for_reply kwarg."""
    await _test_flow_60x(protocol, num_cmds=2)


@pytest.mark.xdist_group(name="virt_serial")
async def test_flow_qos(protocol: PortProtocol) -> None:
    """Check the wait_for_reply kwarg."""
    await _test_flow_qos(protocol)


# @pytest_asyncio.fixture
# async def async_benchmark(benchmark: pytest.FixtureDef) -> Callable[..., None]:
#     event_loop = asyncio.get_running_loop()

#     def _wrapper(func: Callable, *args: Any, **kwargs: Any) -> None:
#         if asyncio.iscoroutinefunction(func):

#             @benchmark
#             def _():
#                 return event_loop.run_until_complete(func(*args, **kwargs))

#         else:
#             benchmark(func, *args, **kwargs)

#     return _wrapper


# @pytest.mark.xdist_group(name="virt_serial")
# def test_benchmark_100(async_benchmark) -> None:
#     async_benchmark(_test_flow_10x)


# @pytest.mark.xdist_group(name="virt_serial")
# def test_benchmark_300(async_benchmark) -> None:
#     async_benchmark(_test_flow_30x)



#!/usr/bin/env python3
"""A CLI for the ramses_rf library."""

from __future__ import annotations

_DBG_CLI = False  # HACK: for debugging of CLI (*before* loading library)


if _DBG_CLI:
    from .debug import start_debugging

    start_debugging(True)



#!/usr/bin/env python3
"""RAMSES RF - a RAMSES-II protocol decoder & analyser."""

import json
import logging
import warnings
from pathlib import Path

from serial.serialutil import SerialException
from serial.tools import list_ports
from tests_deprecated.mocked_rf import CTL_ID, MOCKED_PORT, MockDeviceCtl, MockGateway

from ramses_rf import Gateway
from ramses_rf.schemas import SCH_GLOBAL_CONFIG
from ramses_rf.system import System

warnings.filterwarnings("ignore", category=DeprecationWarning)

logging.disable(logging.WARNING)  # usu. WARNING


TEST_DIR = Path(__file__).resolve().parent

test_ports = {MOCKED_PORT: MockGateway}
if ports := [
    c for c in list_ports.comports() if c.device[-7:-1] in ("ttyACM", "ttyUSB")
]:
    test_ports[ports[0].device] = Gateway

rf_test_failed = False  # global


def abort_if_rf_test_fails(fnc):
    """Abort all non-mocked RF tests once any such test raises a SerialException."""

    async def check_serial_port(test_port, *args, **kwargs):
        if test_port[0] == MOCKED_PORT:
            await fnc(test_port, *args, **kwargs)
            return

        global rf_test_failed
        if rf_test_failed:
            raise SerialException

        try:
            await fnc(test_port, *args, **kwargs)
        except SerialException:
            rf_test_failed = True
            raise

    return check_serial_port


def find_test_tcs(gwy: Gateway) -> System:
    if isinstance(gwy, MockGateway):
        return gwy.system_by_id[CTL_ID]
    systems = [s for s in gwy.systems if s.id != CTL_ID]
    return systems[0] if systems else gwy.system_by_id[CTL_ID]


async def load_test_gwy(
    port_name, gwy_class, config_file: str, devices=None, **kwargs
) -> Gateway:
    """Create a system state from a packet log (using an optional configuration)."""

    config = SCH_GLOBAL_CONFIG({k: v for k, v in kwargs.items() if k[:1] != "_"})

    try:
        with open(config_file) as f:
            config.update(json.load(f))
    except (FileNotFoundError, TypeError):  # TypeError if config_file is None
        pass

    config = SCH_GLOBAL_CONFIG(config)

    gwy: Gateway | MockGateway = gwy_class(port_name, **config)
    await gwy.start(start_discovery=False)  # may: SerialException

    # with patch(
    #     "ramses_tx.transport.serial_for_url",
    #     return_value=MockSerial(gwy.ser_name, loop=gwy._loop),
    # ):
    #     await gwy.start(start_discovery=False)  # may: SerialException

    if hasattr(  # TODO: move out of this routine
        gwy.pkt_transport.serial, "mock_devices"
    ):  # needs ser instance, so after gwy.start()
        gwy.pkt_transport.serial.mock_devices = (
            [MockDeviceCtl(gwy, CTL_ID)] if devices is None else devices
        )

    return gwy



#!/usr/bin/env python3
"""A CLI for the ramses_rf library."""

from __future__ import annotations

import asyncio
import json
import logging
import sys
from typing import Any, Final

import click
from colorama import Fore, Style, init as colorama_init

from ramses_rf import Gateway, GracefulExit, Message, exceptions as exc
from ramses_rf.const import DONT_CREATE_MESSAGES, SZ_ZONE_IDX
from ramses_rf.helpers import deep_merge
from ramses_rf.schemas import (
    SCH_GLOBAL_CONFIG,
    SZ_CONFIG,
    SZ_DISABLE_DISCOVERY,
    SZ_ENABLE_EAVESDROP,
    SZ_REDUCE_PROCESSING,
)
from ramses_tx import is_valid_dev_id
from ramses_tx.logger import CONSOLE_COLS, DEFAULT_DATEFMT, DEFAULT_FMT
from ramses_tx.schemas import (
    SZ_DISABLE_QOS,
    SZ_DISABLE_SENDING,
    SZ_ENFORCE_KNOWN_LIST,
    SZ_EVOFW_FLAG,
    SZ_FILE_NAME,
    SZ_KNOWN_LIST,
    SZ_PACKET_LOG,
    SZ_SERIAL_PORT,
)

from .debug import SZ_DBG_MODE, start_debugging
from .discovery import GET_FAULTS, GET_SCHED, SET_SCHED, spawn_scripts

from ramses_rf.const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    I_,
    RP,
    RQ,
    W_,
    DEV_TYPE_MAP,
    Code,
)

_PROFILE_LIBRARY = False  # NOTE: for profiling of library

if _PROFILE_LIBRARY:
    import cProfile
    import pstats


SZ_INPUT_FILE: Final = "input_file"

# DEFAULT_SUMMARY can be: True, False, or None
SHOW_SCHEMA = False
SHOW_PARAMS = False
SHOW_STATUS = False
SHOW_KNOWNS = False
SHOW_TRAITS = False
SHOW_CRAZYS = False

PRINT_STATE = False  # print engine state
# GET_STATE = False  # get engine state
# SET_STATE = False  # set engine state

# this is called after import colorlog to ensure its handlers wrap the correct streams
logging.basicConfig(level=logging.WARNING, format=DEFAULT_FMT, datefmt=DEFAULT_DATEFMT)


EXECUTE: Final = "execute"
LISTEN: Final = "listen"
MONITOR: Final = "monitor"
PARSE: Final = "parse"


COLORS = {
    I_: Fore.GREEN,
    RP: Fore.CYAN,
    RQ: Fore.CYAN,
    W_: Style.BRIGHT + Fore.MAGENTA,
}

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

LIB_KEYS = tuple(SCH_GLOBAL_CONFIG({}).keys()) + (SZ_SERIAL_PORT,)
LIB_CFG_KEYS = tuple(SCH_GLOBAL_CONFIG({})[SZ_CONFIG].keys()) + (SZ_EVOFW_FLAG,)


def normalise_config(lib_config: dict) -> tuple[str, dict]:
    """Convert a HA config dict into the client library's own format."""

    serial_port = lib_config.pop(SZ_SERIAL_PORT, None)

    # fix for: https://github.com/zxdavb/ramses_rf/issues/96
    packet_log = lib_config.get(SZ_PACKET_LOG)
    if isinstance(packet_log, str):
        packet_log = {SZ_FILE_NAME: packet_log}
    lib_config[SZ_PACKET_LOG] = packet_log

    return serial_port, lib_config


def split_kwargs(obj: tuple[dict, dict], kwargs: dict) -> tuple[dict, dict]:
    """Split kwargs into cli/library kwargs."""
    cli_kwargs, lib_kwargs = obj

    cli_kwargs.update(
        {k: v for k, v in kwargs.items() if k not in LIB_KEYS + LIB_CFG_KEYS}
    )
    lib_kwargs.update({k: v for k, v in kwargs.items() if k in LIB_KEYS})
    lib_kwargs[SZ_CONFIG].update({k: v for k, v in kwargs.items() if k in LIB_CFG_KEYS})

    return cli_kwargs, lib_kwargs


class DeviceIdParamType(click.ParamType):
    name = "device_id"

    def convert(self, value: str, param, ctx):
        if is_valid_dev_id(value):
            return value.upper()
        self.fail(f"{value!r} is not a valid device_id", param, ctx)


# Args/Params for both RF and file
@click.group(context_settings=CONTEXT_SETTINGS)  # , invoke_without_command=True)
@click.option("-z", "--debug-mode", count=True, help="enable debugger")
@click.option("-c", "--config-file", type=click.File("r"))
@click.option("-rk", "--restore-schema", type=click.File("r"), help="from a HA store")
@click.option("-rs", "--restore-state", type=click.File("r"), help=" from a HA store")
@click.option("-r", "--reduce-processing", count=True, help="-rrr will give packets")
@click.option("-lf", "--long-format", is_flag=True, help="dont truncate STDOUT")
@click.option("-e/-ne", "--eavesdrop/--no-eavesdrop", default=None)
@click.option("-g", "--print-state", count=True, help="print state (g=schema, gg=all)")
# @click.option("--get-state/--no-get-state", default=GET_STATE, help="get the engine state")
# @click.option("--set-state/--no-set-state", default=SET_STATE, help="set the engine state")
@click.option(  # show_schema
    "-k/-nk",
    "--show-schema/--no-show-schema",
    default=SHOW_SCHEMA,
    help="display system schema",
)
@click.option(  # show_params
    "-p/-np",
    "--show-params/--no-show-params",
    default=SHOW_PARAMS,
    help="display system params",
)
@click.option(  # show_status
    "-s/-ns",
    "--show-status/--no-show-status",
    default=SHOW_STATUS,
    help="display system state",
)
@click.option(  # show_knowns
    "-n/-nn",
    "--show-knowns/--no-show-knowns",
    default=SHOW_KNOWNS,
    help="display known_list (of devices)",
)
@click.option(  # show_traits
    "-t/-nt",
    "--show-traits/--no-show-traits",
    default=SHOW_TRAITS,
    help="display device traits",
)
@click.option(  # show_crazys
    "-x/-nx",
    "--show-crazys/--no-show-crazys",
    default=SHOW_CRAZYS,
    help="display crazy things",
)
@click.pass_context
def cli(ctx, config_file=None, eavesdrop: None | bool = None, **kwargs):
    """A CLI for the ramses_rf library."""

    if kwargs[SZ_DBG_MODE] > 0:  # Do first
        start_debugging(kwargs[SZ_DBG_MODE] == 1)

    kwargs, lib_kwargs = split_kwargs(({}, {SZ_CONFIG: {}}), kwargs)

    if eavesdrop is not None:
        lib_kwargs[SZ_CONFIG][SZ_ENABLE_EAVESDROP] = eavesdrop

    if config_file:  # TODO: validate with voluptuous, use YAML
        lib_kwargs = deep_merge(
            lib_kwargs, json.load(config_file)
        )  # CLI takes precidence

    ctx.obj = kwargs, lib_kwargs


# Args/Params for packet log only
class FileCommand(click.Command):  # client.py parse <file>
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.params.insert(  # input_file
            0, click.Argument(("input-file",), type=click.File("r"), default=sys.stdin)
        )
        # self.params.insert(  # --packet-log  # NOTE: useful for only for test/dev
        #     1,
        #     click.Option(
        #         ("-o", "--packet-log"),
        #         type=click.Path(),
        #         help="Log all packets to this file",
        #     ),
        # )


# Args/Params for RF packets only
class PortCommand(
    click.Command
):  # client.py <command> <port> --packet-log xxx --evofw3-flag xxx
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.params.insert(0, click.Argument(("serial-port",)))
        """ # self.params.insert(  # --no-discover
        #     1,
        #     click.Option(
        #         ("-d/-nd", "--discover/--no-discover"),
        #         is_flag=True,
        #         default=False,
        #         help="Log all packets to this file",
        #     ),
        # )
        # """
        self.params.insert(  # --packet-log
            2,
            click.Option(
                ("-o", "--packet-log"),
                type=click.Path(),
                help="Log all packets to this file",
            ),
        )
        self.params.insert(  # --evofw-flag
            3,
            click.Option(
                ("-T", "--evofw-flag"),
                type=click.STRING,
                help="Pass this traceflag to evofw",
            ),
        )


#
# 1/4: PARSE (a file, +/- eavesdrop)
@click.command(cls=FileCommand)  # parse a packet log, then stop
@click.pass_obj
def parse(obj, **kwargs):
    """Parse a log file for messages/packets."""
    config, lib_config = split_kwargs(obj, kwargs)

    lib_config[SZ_INPUT_FILE] = config.pop(SZ_INPUT_FILE)

    return PARSE, lib_config, config


#
# 2/4: MONITOR (listen to RF, +/- discovery, +/- eavesdrop)
@click.command(cls=PortCommand)  # (optionally) execute a command/script, then monitor
@click.option("-d/-nd", "--discover/--no-discover", default=None)  # --no-discover
@click.option(  # --exec-cmd 'RQ 01:123456 1F09 00'
    "-x", "--exec-cmd", type=click.STRING, help="e.g. 'RQ 01:123456 1F09 00'"
)
@click.option(  # --execute-scr script device_id
    "-X",
    "--exec-scr",
    type=(str, DeviceIdParamType()),
    help="scan_disc|scan_full|scan_hard|bind device_id",
)
@click.option(  # --poll-devices device_id, device_id,...
    "--poll-devices", type=click.STRING, help="e.g. 'device_id, device_id, ...'"
)
@click.pass_obj
def monitor(obj, discover: None | bool = None, **kwargs):
    """Monitor (eavesdrop and/or probe) a serial port for messages/packets."""
    config, lib_config = split_kwargs(obj, kwargs)

    if discover is None:
        if kwargs["exec_scr"] is None and kwargs["poll_devices"] is None:
            print(" - discovery is enabled")
            lib_config[SZ_CONFIG][SZ_DISABLE_DISCOVERY] = False
        else:
            print(" - discovery is disabled")
            lib_config[SZ_CONFIG][SZ_DISABLE_DISCOVERY] = True

    return MONITOR, lib_config, config


#
# 3/4: EXECUTE (send cmds to RF, +/- discovery, +/- eavesdrop)
@click.command(cls=PortCommand)  # execute a (complex) script, then stop
@click.option("-d/-nd", "--discover/--no-discover", default=None)  # --no-discover
@click.option(  # --exec-cmd 'RQ 01:123456 1F09 00'
    "-x", "--exec-cmd", type=click.STRING, help="e.g. 'RQ 01:123456 1F09 00'"
)
@click.option(  # --get-faults ctl_id
    "--get-faults", type=DeviceIdParamType(), help="controller_id"
)
@click.option(  # --get-schedule ctl_id zone_idx|HW
    "--get-schedule",
    default=[None, None],
    type=(DeviceIdParamType(), str),
    help="controller_id, zone_idx (e.g. '0A', 'HW')",
)
@click.option(  # --set-schedule ctl_id zone_idx|HW
    "--set-schedule",
    default=[None, None],
    type=(DeviceIdParamType(), click.File("r")),
    help="controller_id, filename.json",
)
@click.pass_obj
def execute(obj, **kwargs):
    """Execute any specified scripts, return the results, then quit.

    Disables discovery, and enforces a strict allow_list.
    """
    config, lib_config = split_kwargs(obj, kwargs)

    print(" - discovery is force-disabled")
    lib_config[SZ_CONFIG][SZ_DISABLE_DISCOVERY] = True
    lib_config[SZ_CONFIG][SZ_DISABLE_QOS] = False

    if kwargs[GET_FAULTS]:
        known_list = {kwargs[GET_FAULTS]: {}}
    elif kwargs[GET_SCHED][0]:
        known_list = {kwargs[GET_SCHED][0]: {}}
    elif kwargs[SET_SCHED][0]:
        known_list = {kwargs[SET_SCHED][0]: {}}
    else:
        known_list = {}

    if known_list:
        print(" - known list is force-configured/enforced")
        lib_config[SZ_KNOWN_LIST] = known_list
        lib_config[SZ_CONFIG][SZ_ENFORCE_KNOWN_LIST] = True

    return EXECUTE, lib_config, config


#
# 4/4: LISTEN (to RF, +/- eavesdrop - NO sending/discovery)
@click.command(cls=PortCommand)  # (optionally) execute a command, then listen
@click.pass_obj
def listen(obj, **kwargs):
    """Listen to (eavesdrop only) a serial port for messages/packets."""
    config, lib_config = split_kwargs(obj, kwargs)

    print(" - sending is force-disabled")
    lib_config[SZ_CONFIG][SZ_DISABLE_SENDING] = True

    return LISTEN, lib_config, config


def print_results(gwy: Gateway, **kwargs):
    if kwargs[GET_FAULTS]:
        fault_log = gwy.system_by_id[kwargs[GET_FAULTS]]._faultlog.fault_log

        if fault_log is None:
            print("No fault log, or failed to get the fault log.")
        else:
            [print(f"{k:02X}", v) for k, v in fault_log.items()]

    if kwargs[GET_SCHED][0]:
        system_id, zone_idx = kwargs[GET_SCHED]
        if zone_idx == "HW":
            zone = gwy.system_by_id[system_id].dhw
        else:
            zone = gwy.system_by_id[system_id].zone_by_idx[zone_idx]
        schedule = zone.schedule

        if schedule is None:
            print("Failed to get the schedule.")
        else:
            result = {SZ_ZONE_IDX: zone_idx, "schedule": schedule}
            print(">>> Schedule JSON begins <<<")
            print(json.dumps(result, indent=4))
            print(">>> Schedule JSON ended <<<")

    if kwargs[SET_SCHED][0]:
        system_id, _ = kwargs[GET_SCHED]


def _save_state(gwy: Gateway):
    schema, msgs = gwy.get_state()

    with open("state_msgs.log", "w") as f:
        [f.write(f"{dtm} {pkt}\r\n") for dtm, pkt in msgs.items()]  # if not m._expired

    with open("state_schema.json", "w") as f:
        f.write(json.dumps(schema, indent=4))


def _print_engine_state(gwy: Gateway, **kwargs):
    (schema, packets) = gwy.get_state(include_expired=True)

    if kwargs["print_state"] > 0:
        print(f"schema: {json.dumps(schema, indent=4)}\r\n")
    if kwargs["print_state"] > 1:
        print(f"packets: {json.dumps(packets, indent=4)}\r\n")


def print_summary(gwy: Gateway, **kwargs):
    entity = gwy.tcs or gwy

    if kwargs.get("show_schema"):
        print(f"Schema[{entity}] = {json.dumps(entity.schema, indent=4)}\r\n")

        # schema = {d.id: d.schema for d in sorted(gwy.devices)}
        # print(f"Schema[devices] = {json.dumps({'schema': schema}, indent=4)}\r\n")

    if kwargs.get("show_params"):
        print(f"Params[{entity}] = {json.dumps(entity.params, indent=4)}\r\n")

        params = {d.id: d.params for d in sorted(gwy.devices)}
        print(f"Params[devices] = {json.dumps({'params': params}, indent=4)}\r\n")

    if kwargs.get("show_status"):
        print(f"Status[{entity}] = {json.dumps(entity.status, indent=4)}\r\n")

        status = {d.id: d.status for d in sorted(gwy.devices)}
        print(f"Status[devices] = {json.dumps({'status': status}, indent=4)}\r\n")

    if kwargs.get("show_knowns"):  # show device hints (show-knowns)
        print(f"allow_list (hints) = {json.dumps(gwy._include, indent=4)}\r\n")

    if kwargs.get("show_traits"):  # show device traits
        result = {
            d.id: d.traits  # {k: v for k, v in d.traits.items() if k[:1] == "_"}
            for d in sorted(gwy.devices)
        }
        print(json.dumps(result, indent=4), "\r\n")

    if kwargs.get("show_crazys"):
        for device in [d for d in gwy.devices if d.type == DEV_TYPE_MAP.CTL]:
            for code, verbs in device._msgz.items():
                if code in (Code._0005, Code._000C):
                    for verb in verbs.values():
                        for pkt in verb.values():
                            print(f"{pkt}")
            print()
        for device in [d for d in gwy.devices if d.type == DEV_TYPE_MAP.UFC]:
            for code in device._msgz.values():
                for verb in code.values():
                    for pkt in verb.values():
                        print(f"{pkt}")
            print()


async def async_main(command: str, lib_kwargs: dict, **kwargs):
    """Do certain things."""

    def handle_msg(msg: Message) -> None:
        """Process the message as it arrives (a callback).

        In this case, the message is merely printed.
        """

        if kwargs["long_format"]:  # HACK for test/dev
            print(
                f'{msg.dtm.isoformat(timespec="microseconds")} ... {msg!r}'
                f"  # {msg.payload}"  # or f'  # ("{msg.src!r}", "{msg.dst!r}")'
            )
            return

        dtm = f"{msg.dtm:%H:%M:%S.%f}"[:-3]
        con_cols = CONSOLE_COLS

        if msg.code == Code._PUZZ:
            print(f"{Style.BRIGHT}{Fore.YELLOW}{dtm} {msg}"[:con_cols])
        elif msg.src and msg.src.type == DEV_TYPE_MAP.HGI:
            print(f"{Style.BRIGHT}{COLORS.get(msg.verb)}{dtm} {msg}"[:con_cols])
        elif msg.code == Code._1F09 and msg.verb == I_:
            print(f"{Fore.YELLOW}{dtm} {msg}"[:con_cols])
        elif msg.code in (Code._000A, Code._2309, Code._30C9) and msg._has_array:
            print(f"{Fore.YELLOW}{dtm} {msg}"[:con_cols])
        else:
            print(f"{COLORS.get(msg.verb)}{dtm} {msg}"[:con_cols])

    serial_port, lib_kwargs = normalise_config(lib_kwargs)

    if kwargs["restore_schema"]:
        print(" - restoring client schema from a HA cache...")
        state = json.load(kwargs["restore_schema"])["data"]["client_state"]
        lib_kwargs = lib_kwargs | state["schema"]

    # if serial_port == "/dev/ttyMOCK":
    #     from tests.deprecated.mocked_rf import MockGateway  # FIXME: for test/dev
    #     gwy = MockGateway(serial_port, **lib_kwargs)
    # else:
    gwy = Gateway(serial_port, **lib_kwargs)

    if lib_kwargs[SZ_CONFIG][SZ_REDUCE_PROCESSING] < DONT_CREATE_MESSAGES:
        # library will not send MSGs to STDOUT, so we'll send PKTs instead
        colorama_init(autoreset=True)  # WIP: remove strip=True
        gwy.add_msg_handler(handle_msg)

    if kwargs["restore_state"]:
        print(" - restoring packets from a HA cache...")
        state = json.load(kwargs["restore_state"])["data"]["client_state"]
        await gwy._restore_cached_packets(state["packets"])

    print("\r\nclient.py: Starting engine...")

    try:  # main code here
        await gwy.start()

        # TODO:
        # python client.py -rrr listen /dev/ttyUSB0
        # cat *.log | head | python client.py parse

        if command == EXECUTE:
            tasks = spawn_scripts(gwy, **kwargs)
            await asyncio.gather(*tasks)

        elif command == MONITOR:
            _ = spawn_scripts(gwy, **kwargs)
            await gwy._protocol._wait_connection_lost

        elif command == LISTEN:
            await gwy._protocol._wait_connection_lost

    except asyncio.CancelledError:
        msg = "ended via: CancelledError (e.g. SIGINT)"
    except GracefulExit:
        msg = "ended via: GracefulExit"
    except KeyboardInterrupt:  # FIXME: why isn't this captured here? see main
        msg = "ended via: KeyboardInterrupt"
    except exc.RamsesException as err:
        msg = f"ended via: RamsesException: {err}"
    else:  # if no Exceptions raised, e.g. EOF when parsing, or Ctrl-C?
        msg = "ended without error (e.g. EOF)"
    finally:
        await gwy.stop()

    print(f"\r\nclient.py: Engine stopped: {msg}")

    # if kwargs["save_state"]:
    #    _save_state(gwy)

    if kwargs["print_state"]:
        _print_engine_state(gwy, **kwargs)

    elif command == EXECUTE:
        print_results(gwy, **kwargs)

    print_summary(gwy, **kwargs)


cli.add_command(parse)
cli.add_command(monitor)
cli.add_command(execute)
cli.add_command(listen)


def main() -> None:
    print("\r\nclient.py: Starting ramses_rf...")

    try:
        result = cli(standalone_mode=False)
    except click.NoSuchOption as err:
        print(f"Error: {err}")
        sys.exit(-1)

    if isinstance(result, int):
        sys.exit(result)

    (command, lib_kwargs, kwargs) = result

    if sys.platform == "win32":
        print(" - event_loop_policy set for win32")  # do before asyncio.run()
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    try:
        if _PROFILE_LIBRARY:
            profile = cProfile.Profile()
            profile.run("asyncio.run(main(command, lib_kwargs, **kwargs))")
        else:
            asyncio.run(async_main(command, lib_kwargs, **kwargs))
    except KeyboardInterrupt:  # , SystemExit):
        print("\r\nclient.py: Engine stopped: ended via: KeyboardInterrupt")

    if _PROFILE_LIBRARY:
        ps = pstats.Stats(profile)
        ps.sort_stats(pstats.SortKey.TIME).print_stats(20)

    print(" - finished ramses_rf.\r\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""RAMSES RF - Expose an 0404 schedule (is a stateful process)."""

# TODO: use schemas from evohome_async

from __future__ import annotations

import asyncio
import logging
import struct
import zlib
from collections.abc import Iterable
from datetime import timedelta as td
from typing import TYPE_CHECKING, Any, Final, TypeAlias

import voluptuous as vol  # type: ignore[import, unused-ignore]

from ramses_rf.const import (
    SZ_FRAG_NUMBER,
    SZ_FRAGMENT,
    SZ_SCHEDULE,
    SZ_TOTAL_FRAGS,
    SZ_ZONE_IDX,
)
from ramses_tx.command import Command
from ramses_tx.const import SZ_CHANGE_COUNTER
from ramses_tx.message import Message
from ramses_tx.packet import Packet

from ramses_rf.const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    I_,
    RP,
    RQ,
    W_,
    Code,
)

if TYPE_CHECKING:
    from ramses_rf.system.zones import DhwZone, Zone


_LOGGER = logging.getLogger(__name__)


FIVE_MINS = td(minutes=5)

SZ_MSG: Final = "msg"

SZ_DAY_OF_WEEK: Final = "day_of_week"
SZ_HEAT_SETPOINT: Final = "heat_setpoint"
SZ_SWITCHPOINTS: Final = "switchpoints"
SZ_TIME_OF_DAY: Final = "time_of_day"
SZ_ENABLED: Final = "enabled"

REGEX_TIME_OF_DAY: Final = r"^([0-1][0-9]|2[0-3]):[0-5][05]$"


def schema_sched(schema_switchpoint: vol.Schema) -> vol.Schema:
    schema_sched_day = vol.Schema(
        {
            vol.Required(SZ_DAY_OF_WEEK): int,
            vol.Required(SZ_SWITCHPOINTS): vol.All(
                [schema_switchpoint], vol.Length(min=1)
            ),
        },
        extra=vol.PREVENT_EXTRA,
    )
    return vol.Schema(
        vol.All([schema_sched_day], vol.Length(min=0, max=7)),
        extra=vol.PREVENT_EXTRA,
    )


SCH_SWITCHPOINT_DHW = vol.Schema(
    {
        vol.Required(SZ_TIME_OF_DAY): vol.Match(REGEX_TIME_OF_DAY),
        vol.Required(SZ_ENABLED): bool,
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_SWITCHPOINT_ZON = vol.Schema(
    {
        vol.Required(SZ_TIME_OF_DAY): vol.Match(REGEX_TIME_OF_DAY),
        vol.Required(SZ_HEAT_SETPOINT): vol.All(
            vol.Coerce(float), vol.Range(min=5, max=35)
        ),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_SCHEDULE_DHW = schema_sched(SCH_SWITCHPOINT_DHW)
SCH_SCHEDULE_DHW_FULL = vol.Schema(
    {
        vol.Required(SZ_ZONE_IDX): "HW",
        vol.Required(SZ_SCHEDULE): SCH_SCHEDULE_DHW,
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_SCHEDULE_ZON = schema_sched(SCH_SWITCHPOINT_ZON)
SCH_SCHEDULE_ZON_FULL = vol.Schema(
    {
        vol.Required(SZ_ZONE_IDX): vol.Match(r"0[0-F]"),
        vol.Required(SZ_SCHEDULE): SCH_SCHEDULE_ZON,
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_FULL_SCHEDULE = vol.Schema(
    vol.Any(SCH_SCHEDULE_DHW_FULL, SCH_SCHEDULE_ZON_FULL),
    extra=vol.PREVENT_EXTRA,
)


_PayloadT: TypeAlias = dict[str, Any]  # Message payload
_PayloadSetT: TypeAlias = list[_PayloadT | None]

_SwitchPointT: TypeAlias = dict[str, bool | float | str]
_DayOfWeekT: TypeAlias = dict[str, int | list[_SwitchPointT]]
_ScheduleT: TypeAlias = list[_DayOfWeekT]
_FullScheduleT: TypeAlias = dict[str, str | _ScheduleT]

_FragmentT: TypeAlias = str
_FragmentSetT: TypeAlias = list[_FragmentT]

EMPTY_PAYLOAD_SET: _PayloadSetT = [None]


# TODO: make stateful (a la binding)
class Schedule:  # 0404
    """The schedule of a zone."""

    def __init__(self, zone: DhwZone | Zone) -> None:
        _LOGGER.debug("Schedule(zon=%s).__init__()", zone)

        self.id = zone.id
        self._zone = zone
        self.idx = zone.idx

        self.ctl = zone.ctl
        self.tcs = zone.tcs
        self._gwy = zone._gwy

        self._full_schedule: _FullScheduleT = {}

        self._payload_set: _PayloadSetT = EMPTY_PAYLOAD_SET  # Rx'd
        self._fragments: _FragmentSetT = []  # to Tx

        self._global_ver = 0  # None is a sentinel for 'dont know'
        self._sched_ver = 0  # the global_ver when this schedule was retrieved

    def __str__(self) -> str:
        return f"{self._zone} (schedule)"

    def _handle_msg(self, msg: Message) -> None:
        """Process a schedule packet: if possible, create the corresponding schedule."""

        if msg.code == Code._0006:  # keep up, in cause is useful to know in future
            self._global_ver = msg.payload[SZ_CHANGE_COUNTER]
            return

        if msg.code != Code._0404:
            return

        # can do via here, or via gwy.async_send_cmd(cmd)
        # next line also in self._get_schedule(), so protected here with a lock
        if msg.payload[SZ_TOTAL_FRAGS] != 0xFF and self.tcs.zone_lock_idx != self.idx:
            self._payload_set = self._update_payload_set(self._payload_set, msg.payload)

    async def _is_dated(self, *, force_io: bool = False) -> tuple[bool, bool]:
        """Indicate if it is possible that a more recent schedule is available.

        If required, retrieve the latest global version (change counter) from the
        TCS.

        There may be a false positive if another zone's schedule is changed when
        this zone's schedule has not. There may be a false negative if this zone's
        schedule was changed only very recently and a cached global version was
        used.

        If `force_io`, then a true negative is guaranteed (it forces an RQ|0006 unless
        self._global_ver > self._sched_ver).
        """

        # this will not cause an I/O...
        if (
            not force_io
            and not self._sched_ver
            or (self._global_ver and self._global_ver > self._sched_ver)
        ):
            return True, False  # is_dated, did_io

        # this may cause an I/O...
        self._global_ver, did_io = await self.tcs._schedule_version()
        if did_io or self._global_ver > self._sched_ver:
            return self._global_ver > self._sched_ver, did_io  # is_dated, did_io

        if force_io:  # this will cause an I/O...
            self._global_ver, did_io = await self.tcs._schedule_version(
                force_io=force_io
            )

        return self._global_ver > self._sched_ver, did_io  # is_dated, did_io

    async def get_schedule(
        self, *, force_io: bool = False, timeout: float = 15
    ) -> _ScheduleT | None:
        """Retrieve/return the brief schedule of a zone.

        Return the cached schedule (which may have been eavesdropped) only if the
        global change counter has not increased.
        Otherwise, RQ the latest schedule from the controller and return that.

        If `force_io`, then the latest schedule is guaranteed (it forces an RQ|0006).
        """

        try:
            await asyncio.wait_for(
                self._get_schedule(force_io=force_io), timeout=timeout
            )
        except TimeoutError as err:
            raise TimeoutError(
                f"Failed to obtain schedule within {timeout} secs"
            ) from err
        # TODO: raise a more parochial exception
        return self.schedule

    async def _get_schedule(self, *, force_io: bool = False) -> None:
        """Retrieve/return the schedule of a zone (sets self._full_schedule)."""

        async def get_fragment(frag_num: int) -> _PayloadT:  # may: TimeoutError?
            """Retrieve a schedule fragment from the controller."""

            frag_set_size = 0 if frag_num == 1 else _len(self._payload_set)
            cmd = Command.get_schedule_fragment(
                self.ctl.id, self.idx, frag_num, frag_set_size
            )
            pkt: Packet = await self._gwy.async_send_cmd(cmd, wait_for_reply=True)
            msg = Message(pkt)
            assert isinstance(msg.payload, dict)  # mypy check
            return msg.payload  # may: TimeoutError?

        is_dated, did_io = await self._is_dated(force_io=force_io)
        if is_dated:
            self._full_schedule = {}  # keep fragments, maybe only other sched(s) changed
        if self._full_schedule:
            return

        await self.tcs._obtain_lock(self.idx)  # maybe raise TimeOutError

        if not did_io:  # must know the version of the schedule about to be RQ'd
            self._global_ver, _ = await self.tcs._schedule_version(force_io=True)

        self._payload_set[0] = None  # if 1st frag valid: schedule very likely unchanged
        while frag_num := next(
            i for i, f in enumerate(self._payload_set, 1) if f is None
        ):
            fragment = await get_fragment(frag_num)
            # next line also in self._handle_msg(), so protected there with a lock
            self._payload_set = self._update_payload_set(self._payload_set, fragment)
            if self._full_schedule:  # TODO: potential for infinite loop?
                self._sched_ver = self._global_ver
                break

        self.tcs._release_lock()

    def _proc_payload_set(self, payload_set: _PayloadSetT) -> _FullScheduleT | None:
        """Process a payload set and return the full schedule (sets `self._schedule`).

        If the schedule is for DHW, set the `zone_idx` key to 'HW' (to avoid confusing
        with zone '00').
        """

        # TODO: relying upon caller to ensure set is only empty or full

        if payload_set == EMPTY_PAYLOAD_SET:
            self._full_schedule = {SZ_ZONE_IDX: self.idx}
            return self._full_schedule

        try:
            schedule = fragz_to_full_sched(
                payload[SZ_FRAGMENT] for payload in payload_set if payload
            )  # TODO: messy - what is set not full
        except zlib.error:
            return None  # TODO: raise a more parochial exception

        if self.idx == "HW":
            schedule[SZ_ZONE_IDX] = "HW"
        self._full_schedule = schedule

        return self._full_schedule  # NOTE: not self.schedule

    def _update_payload_set(
        self, payload_set: _PayloadSetT, payload: _PayloadT
    ) -> _PayloadSetT:
        """Add a fragment to a frag set and process/return the new set.

        If the frag set is complete, check for a schedule (sets `self._schedule`).

        If required, start a new frag set with the fragment.
        """

        def init_payload_set(payload: _PayloadT) -> _PayloadSetT:
            payload_set: _PayloadSetT = [None] * payload[SZ_TOTAL_FRAGS]
            payload_set[payload[SZ_FRAG_NUMBER] - 1] = payload
            return payload_set

        if payload[SZ_TOTAL_FRAGS] is None:  # zone has no schedule
            payload_set = EMPTY_PAYLOAD_SET
            self._proc_payload_set(payload_set)
            return payload_set

        if payload[SZ_TOTAL_FRAGS] != _len(payload_set):  # sched has changed
            return init_payload_set(payload)

        payload_set[payload[SZ_FRAG_NUMBER] - 1] = payload
        if None in payload_set or self._proc_payload_set(
            payload_set
        ):  # sets self._schedule
            return payload_set

        return init_payload_set(payload)

    async def set_schedule(
        self, schedule: _ScheduleT, force_refresh: bool = False
    ) -> _ScheduleT | None:
        """Set the schedule of a zone."""

        async def put_fragment(frag_num: int, frag_cnt: int, fragment: str) -> None:
            """Send a schedule fragment to the controller."""

            cmd = Command.set_schedule_fragment(
                self.ctl.id, self.idx, frag_num, frag_cnt, fragment
            )
            await self._gwy.async_send_cmd(cmd, wait_for_reply=True)

        def normalise_validate(schedule: _ScheduleT) -> _FullScheduleT:
            full_schedule: _FullScheduleT

            if self.idx == "HW":
                full_schedule = {SZ_ZONE_IDX: "HW", SZ_SCHEDULE: schedule}
                schedule_schema = SCH_SCHEDULE_DHW_FULL
            else:
                full_schedule = {SZ_ZONE_IDX: self.idx, SZ_SCHEDULE: schedule}
                schedule_schema = SCH_SCHEDULE_ZON_FULL

            try:
                full_schedule = schedule_schema(full_schedule)
            except vol.MultipleInvalid as err:
                raise TypeError(f"failed to set schedule: {err}") from err

            if self.idx == "HW":  # HACK: to avoid confusing dhw with zone '00'
                full_schedule[SZ_ZONE_IDX] = "00"

            return full_schedule

        full_schedule = normalise_validate(schedule)
        self._fragments = full_sched_to_fragz(full_schedule)

        await self.tcs._obtain_lock(self.idx)  # maybe raise TimeOutError

        try:
            for num, frag in enumerate(self._fragments, 1):
                await put_fragment(num, len(self._fragments), frag)
        except TimeoutError as err:
            raise TimeoutError(f"failed to set schedule: {err}") from err
        else:
            if not force_refresh:
                self._global_ver, _ = await self.tcs._schedule_version(force_io=True)
                # assert self._global_ver > self._sched_ver
                self._sched_ver = self._global_ver
        finally:
            self.tcs._release_lock()

        if force_refresh:
            await self.get_schedule(force_io=True)  # sets self._full_schedule
        else:
            self._full_schedule = full_schedule

        return self.schedule

    @property
    def schedule(self) -> _ScheduleT | None:
        """Return the current (not full) schedule, if any."""
        if not self._full_schedule:  # can be {}
            return None
        result: _ScheduleT = self._full_schedule.get(SZ_SCHEDULE)  # type: ignore[assignment]
        return result

    @property
    def version(self) -> int | None:
        """Return the version associated with the current schedule, if any."""
        return self._sched_ver if self._full_schedule else None


# TODO: deprecate in favour of len(payload_set)
def _len(payload_set: _PayloadSetT) -> int:
    """Return the total number of fragments in the complete frag set.

    Return 0 if the expected set size is unknown (sentinel value as per RAMSES II).

    Uses frag_set[i][SZ_TOTAL_FRAGS] instead of `len(frag_set)` (is necessary?).
    """
    # for frag in (f for f in payload_set if f is not None):  # they will all match
    #     assert len(payload_set) == frag[SZ_TOTAL_FRAGS]  # TODO: remove
    #     assert isinstance(frag[SZ_TOTAL_FRAGS], int)  # mypy check
    #     result: int = frag[SZ_TOTAL_FRAGS]
    #     return result

    # assert payload_set == EMPTY_PAYLOAD_SET  # TODO: remove
    # return 0  # sentinel value as per RAMSES protocol
    return len(payload_set)


def fragz_to_full_sched(fragments: Iterable[_FragmentT]) -> _FullScheduleT:
    """Convert a tuple of fragments strs (a blob) into a schedule.

    May raise a `zlib.error` exception.
    """

    raw_schedule = zlib.decompress(bytearray.fromhex("".join(fragments)))

    old_day = 0
    schedule: _ScheduleT = []
    switchpoints: list[_SwitchPointT] = []

    idx: int
    dow: int
    tod: int
    val: int

    for i in range(0, len(raw_schedule), 20):
        idx, dow, tod, val = _struct_unpack(raw_schedule[i : i + 20])

        if dow > old_day:
            schedule.append({SZ_DAY_OF_WEEK: old_day, SZ_SWITCHPOINTS: switchpoints})
            old_day, switchpoints = dow, []

        switchpoints.append(
            {
                SZ_TIME_OF_DAY: "{:02d}:{:02d}".format(*divmod(tod, 60)),
                **(
                    {SZ_ENABLED: bool(val)}
                    if val in (0, 1)
                    else {SZ_HEAT_SETPOINT: val / 100}
                ),
            }
        )

    schedule.append({SZ_DAY_OF_WEEK: old_day, SZ_SWITCHPOINTS: switchpoints})

    return {SZ_ZONE_IDX: f"{idx:02X}", SZ_SCHEDULE: schedule}


def full_sched_to_fragz(full_schedule: _FullScheduleT) -> list[_FragmentT]:
    """Convert a schedule into a set of fragments (a blob).

    May raise `KeyError`, `zlib.error` exceptions.
    """

    cobj = zlib.compressobj(level=9, wbits=14)
    frags: list[bytes] = []

    week_days: _ScheduleT = full_schedule[SZ_SCHEDULE]  # type: ignore[assignment]
    for week_day in week_days:
        switchpoints: list[_SwitchPointT] = week_day[SZ_SWITCHPOINTS]  # type: ignore[assignment]
        for switchpoint in switchpoints:
            frags.append(_struct_pack(full_schedule, week_day, switchpoint))

    blob = (b"".join(cobj.compress(f) for f in frags) + cobj.flush()).hex().upper()

    return [blob[i : i + 82] for i in range(0, len(blob), 82)]


def _struct_pack(
    full_schedule: _FullScheduleT, week_day: _DayOfWeekT, switchpoint: _SwitchPointT
) -> bytes:
    idx_: str = full_schedule[SZ_ZONE_IDX]  # type: ignore[assignment]
    dow_: int = week_day[SZ_DAY_OF_WEEK]  # type: ignore[assignment]
    tod_: str = switchpoint[SZ_TIME_OF_DAY]  # type: ignore[assignment]

    idx = int(idx_, 16)
    dow = int(dow_)
    tod = int(tod_[:2]) * 60 + int(tod_[3:])

    if SZ_HEAT_SETPOINT in switchpoint:
        val = int(switchpoint[SZ_HEAT_SETPOINT] * 100)
    else:
        val = int(bool(switchpoint[SZ_ENABLED]))

    return struct.pack("<xxxxBxxxBxxxHxxHxx", idx, dow, tod, val)


def _struct_unpack(raw_schedule: bytes) -> tuple[int, int, int, int]:
    idx, dow, tod, val, _ = struct.unpack("<xxxxBxxxBxxxHxxHH", raw_schedule)
    return idx, dow, tod, val


# 16:27:56.942 000 RQ --- 18:006402 01:145038 --:------ 0006 001 00
# 16:27:56.958 038 RP --- 01:145038 18:006402 --:------ 0006 004 00050009

# 16:27:57.005 000 RQ --- 18:006402 01:145038 --:------ 0404 007 0120000800-0100
# 16:27:57.068 037 RP --- 01:145038 18:006402 --:------ 0404 048 0120000829-0103-68816DCFCB0980301045D1994C3E624916660956604596600516E1D285094112F566F5B80C072222A2
# 16:27:57.114 000 RQ --- 18:006402 01:145038 --:------ 0404 007 0120000800-0203
# 16:27:57.161 038 RP --- 01:145038 18:006402 --:------ 0404 048 0120000829-0203-52DF92C79CEA7EDA91C7F06997FDEFC620B287D6143C054FC153F01C780E3C079E03CFC033F00C3C03
# 16:27:57.202 000 RQ --- 18:006402 01:145038 --:------ 0404 007 0120000800-0303
# 16:27:57.245 038 RP --- 01:145038 18:006402 --:------ 0404 045 0120000826-0303-CF83E7C1F3E079F0CADC3E5E696BFECC944EED5BF5DEAD7AAD45F0227811BCD87937936E24CF



#!/usr/bin/env python3
"""RAMSES RF - Expose an 0418 fault log (is a stateful process)."""

from __future__ import annotations

import dataclasses
import logging
from collections import OrderedDict
from typing import TYPE_CHECKING, Never, NewType, TypeAlias

from ramses_tx import Command, Message, Packet
from ramses_tx.const import (
    SZ_LOG_ENTRY,
    SZ_LOG_IDX,
    FaultDeviceClass,
    FaultState,
    FaultType,
)
from ramses_tx.helpers import parse_fault_log_entry
from ramses_tx.schemas import DeviceIdT

from ramses_rf.const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    I_,
    RP,
    RQ,
    W_,
    Code,
)

if TYPE_CHECKING:
    from ramses_rf.system.heat import Evohome


_LOGGER = logging.getLogger(__name__)


#  {'log_idx': '00', 'log_entry': ('21-12-23T11:59:35', 'restore', 'battery_low', 'actuator', '00', '04:164787', 'B0', '0000', 'FFFF7000')}


@dataclasses.dataclass(frozen=True, kw_only=True, order=True)
class FaultLogEntry:
    """A fault log entry of an evohome fault log.

    Fault log entries do have a log_idx attr, but this is merely their current location
    in the system's fault log.
    """

    timestamp: str  # #               # 21-12-23T11:59:35 - assume is unique
    fault_state: FaultState  # #      # fault, restore, unknown_c0
    fault_type: FaultType  # #        # system_fault, battery_low, sensor_fault, etc.
    domain_idx: str  # #              # 00-0F, FC, etc. ? only if dev_class is/not CTL?
    device_class: FaultDeviceClass  # # controller, actuator, sensor, etc.
    device_id: DeviceIdT | None  # #  # 04:164787

    @classmethod
    def from_msg(cls, msg: Message) -> FaultLogEntry:
        """Create a fault log entry from a message's packet."""
        return cls.from_pkt(msg._pkt)

    @classmethod
    def from_pkt(cls, pkt: Packet) -> FaultLogEntry:
        """Create a fault log entry from a packet's payload."""

        log_entry = parse_fault_log_entry(pkt.payload)
        if log_entry is None:
            raise TypeError("Null fault log entry")

        return cls(**{k: v for k, v in log_entry.items() if k[:1] != "_"})  # type: ignore[arg-type]


FaultDtmT = NewType("FaultDtmT", str)
FaultIdxT = NewType("FaultIdxT", int)

FaultLogT: TypeAlias = dict[Never, Never] | dict[FaultDtmT, FaultLogEntry]
FaultMapT: TypeAlias = OrderedDict[Never, Never] | OrderedDict[FaultIdxT, FaultDtmT]


class FaultLog:  # 0418  # TODO: use a NamedTuple
    """The fault log of an evohome system.

    This code assumes that the `timestamp` attr of each log entry is a unique identifer.

    Null entries do not have a timestamp. All subsequent entries will also be null.

    The `log_idx` is not an identifier: it is merely the current position of a log entry
    in the system log.

    New entries are added to the top of the log (log_idx=0), and the log_idx is
    incremented for all exisiting log enties.
    """

    _MAX_LOG_IDX = 0x3E

    def __init__(self, tcs: Evohome) -> None:
        self._tcs = tcs
        self.id = tcs.id
        self._gwy = tcs._gwy

        self._log: FaultLogT = dict()
        self._map: FaultMapT = OrderedDict()
        self._log_done: bool | None = None

        self._is_current: bool = False  # if we now our log is out of date
        self._is_getting: bool = False

    def _insert_into_map(self, idx: FaultIdxT, dtm: FaultDtmT | None) -> FaultMapT:
        """Rebuild the map, given the new log entry data."""

        new_map: FaultMapT = OrderedDict()

        # usu. idx == 0, but could be > 0
        new_map |= {
            k: v for k, v in self._map.items() if k < idx and (dtm is None or v > dtm)
        }

        if dtm is None:  # there are no subsequent log entries
            return new_map

        new_map |= {idx: dtm}

        if not (idxs := [k for k, v in self._map.items() if v < dtm]):
            return new_map

        if (next_idx := min(idxs)) > idx:
            diff = 0
        elif next_idx == idx:
            diff = 1  # next - idx + 1
        else:
            diff = idx + 1  # 1 if self._map.get(idx) else 0

        new_map |= {
            k + diff: v  # type: ignore[misc]
            for k, v in self._map.items()
            if (k >= idx or v < dtm) and k + diff <= self._MAX_LOG_IDX
        }

        return new_map

    def _handle_msg(self, msg: Message) -> None:
        """Handle a fault log message."""

        if msg.code != Code._0418:
            return

        if msg.verb == I_:
            self._is_current = False

        if SZ_LOG_IDX in msg.payload:
            idx: FaultIdxT = int(msg.payload[SZ_LOG_IDX], 16)  # type: ignore[assignment]
        elif msg._pkt._idx:  # then idx was hacked in by protocol FSM
            idx = int(msg._pkt._idx, 16)  # type: ignore[assignment, arg-type]
        else:
            return  # we can't do anything useful with this message

        if msg.payload[SZ_LOG_ENTRY] is None:  # NOTE: Subsequent entries will be empty
            self._map = self._insert_into_map(idx, None)
            self._log = {k: v for k, v in self._log.items() if k in self._map.values()}
            return  # If idx != 0, should we also check from idx = 0?

        entry = FaultLogEntry.from_msg(msg)  # if msg.payload[SZ_LOG_ENTRY] else None
        dtm: FaultDtmT = entry.timestamp  # type: ignore[assignment]

        if self._map.get(idx) == dtm:  # type: ignore[call-overload]
            return  # i.e. No evidence anything has changed

        if dtm not in self._log:
            self._log |= {dtm: entry}  # must add entry before _insert_into_map()
        self._map = self._insert_into_map(idx, dtm)  # updates self._map
        self._log = {k: v for k, v in self._log.items() if k in self._map.values()}

        # if idx != 0:  # there's other (new/changed) entries above this one?
        #     pass

    async def get_faultlog(
        self,
        /,
        *,
        start: int = 0,
        limit: int | None = None,
        force_refresh: bool = False,
    ) -> dict[FaultIdxT, FaultLogEntry]:
        """Retrieve the fault log from the controller."""
        if limit is None:
            limit = self._MAX_LOG_IDX + 1

        self._is_getting = True

        for i in range(start, limit):
            cmd = Command.get_system_log_entry(self.id, i)
            pkt = await self._gwy.async_send_cmd(cmd, wait_for_reply=True)

            try:
                _ = FaultLogEntry.from_pkt(pkt)
            except TypeError:  # Null fault log entry
                self._handle_msg(Message(pkt))
                break
            except AttributeError:
                break

        self._is_current = False
        self._is_getting = False

        return self.faultlog

    @property
    def faultlog(self) -> dict[FaultIdxT, FaultLogEntry]:
        """Return the fault log of a system."""

        # if self._faultlog:
        #     return self._faultlog

        return {idx: self._log[dtm] for idx, dtm in self._map.items()}  # type: ignore[index]

    def is_current(self, force_io: bool | None = None) -> bool:
        """Return True if the local fault log is identical to the controllers.

        If force_io, retrieve the 0th log entry and check it is identical to the local
        copy.
        """

        if not self._is_current:
            return False
        return True



#!/usr/bin/env python3
"""RAMSES RF - Check get of TCS fault logs."""

import asyncio

import pytest

from ramses_rf import Gateway
from ramses_rf.device import Controller
from ramses_rf.system import Evohome
from ramses_tx.address import HGI_DEVICE_ID
from ramses_tx.protocol import PortProtocol
from ramses_tx.schemas import DeviceIdT

from .conftest import _GwyConfigDictT

# ### FIXTURES #########################################################################

pytestmark = pytest.mark.asyncio()


@pytest.fixture()
def gwy_config() -> _GwyConfigDictT:
    return {
        "config": {
            "disable_discovery": True,
            "disable_qos": False,  # QoS is required for this test
            "enforce_known_list": False,
        },
        "known_list": {HGI_DEVICE_ID: {}},  # req'd to thwart foreign HGI blacklisting
    }


#######################################################################################


async def _test_get_faultlog(gwy: Gateway, ctl_id: DeviceIdT) -> None:
    """Test obtaining the fault log."""

    assert gwy._loop is asyncio.get_running_loop()  # scope BUG is here
    assert isinstance(gwy._protocol, PortProtocol)  # mypy
    assert gwy._protocol._disable_qos is False  # QoS is required for this test

    _: Controller = gwy.get_device(ctl_id)  # type: ignore[assignment]

    tcs: Evohome | None = gwy.tcs
    assert isinstance(tcs, Evohome)  # mypy

    faultlog = await tcs.get_faultlog()
    assert faultlog


#######################################################################################


@pytest.mark.xdist_group(name="real_serial")
async def test_get_faultlog_mqtt(mqtt_evofw3: Gateway) -> None:
    """Test obtaining the fault log from a real controller via MQTT."""

    await _test_get_faultlog(mqtt_evofw3, "01:145038")


@pytest.mark.xdist_group(name="real_serial")
async def test_get_faultlog_real(real_evofw3: Gateway) -> None:
    """Test obtaining the fault log from a real controller via RF."""

    await _test_get_faultlog(real_evofw3, "01:145038")
